﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Text.RegularExpressions;
using Ordinamento;
using Quine_McCluskey;

namespace Algoritmo_Quine_McCluskey
{
    public partial class Main : Form
    {
        //Property: tmedc_reg
        //Filtro per l'input in textbox (sono ammessi solo numeri e virgola)
        public Regex tmedc_reg = new Regex("[^0-9,]");
        //Variable: termini_minimi
        //I termini minimi della funzione dati in input dall'utente.
        public int[] termini_minimi;
        //Variable: dontcare
        //Le condizioni di indifferenza date in input dall'utente.
        public int[] dontcare;
        //Variable: numero_variabili
        //Il numero di variabili della funzione
        public int numero_variabili = 2;
        //Variable: ins_tm_eff
        //Variabile booleana che indica l'avvenuto, e corretto, inserimento dei termini minimi
        public bool ins_tm_eff = false; 

        public List<string> dontcare_s = new List<string>(), termini_minimi_s = new List<string>();
        //Variable: qm
        //Viene utilizzata la classe Quine-McCluskey per la riduzione
        public QuineMcClusky qm = new QuineMcClusky(); //La classe QuineMcCluskey ha una documentazione a se stante
        public Main()
        {
            InitializeComponent();
        }

        //Function: dontcare_TBox_TextChanged
        //Questa procedura viene chiamata ogni qualvolta
        //l'utente scrive sulla textbox relativa alle condizioni di indifferenza
        private void dontcare_TBox_TextChanged(object sender, EventArgs e)
        {
            dontcare_TBox.Text = tmedc_reg.Replace(dontcare_TBox.Text, ""); //Elimina tutti i caratteri non ammessi dal filtro
            dontcare_TBox.Focus(); //Riporta il focus sulla textbox rendendolo l'elemento attivo nel form
            dontcare_TBox.Select(dontcare_TBox.Text.Length, 0); //Riporta il cursore alla fine del testo nella textbox.
            //Questo perchè le procedure precendenti lo spostano all'inizio.
        }

        //Function: termini_minimi_TBox_TextChanged
        //Stessa cosa del precedente per i termini minimi
        private void termini_minimi_TBox_TextChanged(object sender, EventArgs e)
        {
            termini_minimi_TBox.Text = tmedc_reg.Replace(termini_minimi_TBox.Text, "");
            termini_minimi_TBox.Focus();
            termini_minimi_TBox.Select(termini_minimi_TBox.Text.Length, 0);
        }

        //Function: termini_minimi_TBox_Leave
        //Questa procedura viene chiamata nel momento in cui la textbox dei termini minimi smette di essere l'elemento attivo del form
        private void termini_minimi_TBox_Leave(object sender, EventArgs e)
        {
            Errore_status.Hide(); //Viene nascosto il messaggio sull'input ammesso

            for (int i = 0; i < termini_minimi_TBox.Text.Length - 1; i++) //Elimina le virgole duplicate nella textbox
            {
                if (termini_minimi_TBox.Text[i] == ',' && termini_minimi_TBox.Text[i + 1] == ',')
                {
                    termini_minimi_TBox.Text = termini_minimi_TBox.Text.Remove(i, 1);
                    i--;
                }
            }

            if(termini_minimi_TBox.Text.StartsWith(",")) //Qualora il testo nella textbox inizi con una virgola invece che
            //con un numero questa viene eliminata
            {
                termini_minimi_TBox.Text = termini_minimi_TBox.Text.Remove(0, 1);
            }

            if (String.IsNullOrEmpty(termini_minimi_TBox.Text)) //Se non sono stati inseriti termini minimi
            {
                ins_tm_eff = false; //La variabile ins_tm_eff è impostata su false
            }

            else //Se sono stati inseriti dei termini minimi
            {
                termini_minimi_s = termini_minimi_TBox.Text.Split(',').ToList<string>(); //La lista termini_minimi_s
                //viene riempita con i termini minimi inseriti nella textbox
         
                bool doppioni = false; //Variabile che serve ad indicare se siano o meno stati inseriti termini doppioni

                bool rimozione_effettuata = false; //Variabile che serve a verificare che sono stati fatti degli aggiustamenti
                //sull'input dell'utente rimuovendo termini non consentiti. Il suo valore verrà analizzato
                //per mostrare a video eventualmente un messaggio.

                string nuovo_testo = ""; //Stringa che sostituirà il testo nella textbox

                termini_minimi_s.Sort(); //Ordina la lista alfabeticamente

                int indice = 0; 
                while (indice < termini_minimi_s.Count - 1) //Procedura di rimozione doppioni dalla lista
                {
                    if (termini_minimi_s[indice] == termini_minimi_s[indice + 1])
                    {
                        termini_minimi_s.RemoveAt(indice);
                        doppioni = true;
                    }
                    else
                        indice++;
                }

                if (doppioni == true) //Se sono stati rimossi dei doppioni
                {
                    foreach (string s in termini_minimi_s) //Reimposta il testo nella textbox togliendo i doppioni
                    {
                        nuovo_testo += s; 
                        nuovo_testo += ",";
                    }
                    nuovo_testo = nuovo_testo.Remove(nuovo_testo.Length - 1, 1); //Rimuove l'ultima virgola della stringa

                    MessageBox.Show(this, "Sono stati rimossi i termini minimi doppioni inseriti."); //Messaggio che informa
                    //dell'avvenuta rimozione di doppioni

                    termini_minimi_TBox.Text = nuovo_testo;
                }

                doppioni = false; //Viene reimpostata la variabile doppioni su false, poiché verrà riutilizzata
                                  //nella verifica della presenza di termini minimi presenti anche tra le condizioni di indifferenza

                if (dontcare_TBox.Text != "")
                {
                    for (indice = 0; indice < termini_minimi_s.Count; indice++) //Per ogni termine minimo
                    {
                        foreach (string s in dontcare_s) //Controlla che non sia presenta anche tra le condizioni di indifferenza
                        {
                            if (termini_minimi_s[indice] == s) //In tal caso
                            {
                                termini_minimi_s.Remove(termini_minimi_s[indice]); //Lo rimuove
                                doppioni = true; //E imposta la variabile doppioni su true
                            }
                        }
                    }
                }


                if (doppioni == true)
                {
                    nuovo_testo = "";
                    foreach (string s in termini_minimi_s)
                    {
                        nuovo_testo += s;
                        nuovo_testo += ",";
                    }
                    nuovo_testo = nuovo_testo.Remove(nuovo_testo.Length - 1, 1);
                    MessageBox.Show(this, "Sono stati rimossi i termini minimi ritrovati anche tra le condizioni di indifferenza.");
                    termini_minimi_TBox.Text = nuovo_testo;
                }

                int n, k = 0;
                int numero_elementi = termini_minimi_s.Count;

                for (int i = 0; i < termini_minimi_s.Count; i++) //Verifica che i termini inseriti dall'utente rientrino nel
                //range definito dal numero di variabili
                {
                    if (Int32.TryParse(termini_minimi_s[i], out n)) //Procedura sicura di conversione dell'input in forma numerica
                    {
                        if (n > (int)Math.Pow(2, numero_variabili) - 1) //Se il singolo termine è maggiore di 2 elevato al
                        //numero di variabili meno uno allora è fuori range e viene rimosso
                        {
                            termini_minimi_s.Remove(termini_minimi_s[i]);
                            i--;
                            k++;
                            rimozione_effettuata = true; //E la variabile rimozione_effettuata viene impostata su true
                        }
                    }
                }

                if (rimozione_effettuata == true && numero_elementi == k) //Se tutti i termini sono stati rimossi
                //perchè fuori range
                {
                    termini_minimi_TBox.Text = ""; //Pulisce la textbox
                    //e mostra il messaggio seguente
                    MessageBox.Show(this, "Sono state effettuate delle rimozioni di termini minimi non consentiti. Controllare il numero di variabili", "Fuori Range", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                }

                else
                {
                    if (rimozione_effettuata == true) //Se invece sono stati rimossi solo alcuni dei termini
                    {
                        //mostra questo messaggio
                        MessageBox.Show(this, "Sono state effettuate delle rimozioni di termini minimi non consentiti. Controllare il numero di variabili", "Fuori Range", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                    }

                    //e procede alla riscrittura del testo nella textbox
                    nuovo_testo = ""; //Il nuovo testo che sostituirà il precedente
                    Confronta_Termini ct = new Confronta_Termini(); //Procedura di confronto per ordinare i termini dal più
                    //piccolo al più grande
                    termini_minimi_s.Sort(0, termini_minimi_s.Count, ct); //I termini vengono ordinati secondo la procedura
                    //Confronta_Termini()


                    foreach (string s in termini_minimi_s) //Si riempe la stringa nuovo_testo con i termini minimi rimasti
                    {
                        nuovo_testo += s;
                        nuovo_testo += ",";
                    }
                    nuovo_testo = nuovo_testo.Remove(nuovo_testo.Length - 1, 1); //Rimuove l'ultima virgola che era stata inserita

                    termini_minimi_TBox.Text = nuovo_testo;
                    ins_tm_eff = true; //Adesso con certezza i termini minimi sono tutti validi e pertanto l'inserimento
                    //è stato effettuato
                }
            }
        }

        private void Main_Load(object sender, EventArgs e) //Procedura al caricamento della finestra
        {
            n_var_cBox.SelectedItem = "2"; //Seleziona 2 come numero di variabili
            Errore_status.Hide(); //Nasconde il messaggio che indica i valori consentiti
        }

        private void step_successivo_button_Click(object sender, EventArgs e) //Alla pressione del bottone Step Successivo
        {
            if (fine_checkbox.Checked == true) //Come in primo_step_button_Click
            {
                while (qm.fine != true)
                    qm.Step_Successivo();
                output.Text += "\nForma minima della funzione ottenuta tramite l'algoritmo di Quine Mc-Cluskey:\n";
                foreach (string s in qm.Stampa_Stato_Finale())
                {
                    output.Text += s;
                }

                output.Text += "\n\n";

                if (petrick_checkbox.Checked == true)
                {
                    foreach (string s in qm.Stampa_Stato_Finale_Petrick())
                    {
                        output.Text += s;
                    }
                }
                step_successivo_button.Enabled = false;
                azzera_button.Enabled = true;
                output.Focus();
                output.Select(output.TextLength, 0);
                fine_checkbox.Checked = false;
                petrick_checkbox.Checked = false;
            }
            else
            {
                qm.Step_Successivo();
                if (qm.fine == false)
                {
                    foreach (string s in qm.Stampa_Stato_Corrente())
                    {
                        output.Text += s;
                    }
                    output.Text += "\n\n";
                    output.Focus();
                    output.Select(output.TextLength, 0);
                }
                else
                {
                    output.Text += "\nForma minima della funzione ottenuta tramite l'algoritmo di Quine Mc-Cluskey:\n";
                    foreach (string s in qm.Stampa_Stato_Finale())
                    {
                        output.Text += s;
                    }

                    output.Text += "\n\n";

                    if (petrick_checkbox.Checked == true)
                    {
                        foreach (string s in qm.Stampa_Stato_Finale_Petrick())
                        {
                            output.Text += s;
                        }
                    }
                    step_successivo_button.Enabled = false;
                    azzera_button.Enabled = true;
                    output.Focus();
                    output.Select(output.TextLength, 0);
                }
            }
        }

        private void n_var_cBox_SelectedIndexChanged(object sender, EventArgs e) //Quando cambia il numero di variabili
        //selezionato
        {
            numero_variabili = n_var_cBox.SelectedIndex + 2;
        }

        private void copia_toolstrip_Click(object sender, EventArgs e)
        {
            if(output.SelectedText != "")
                Clipboard.SetText(output.SelectedText);
        }

        private void termini_minimi_TBox_Enter(object sender, EventArgs e) //All'ingresso nella textbox termini minimi
        {
            Errore_status.Show(); //Viene mostrato il messaggio che indica l'input ammesso
        }

        private void dontcare_TBox_Enter(object sender, EventArgs e) //All'ingresso nella textbox delle condizioni di indifferenza
        {
            Errore_status.Show(); //Viene mostrato il messaggio che indica l'input ammesso
        }

        //Function: dontcare_TBox_Leave
        //Procedura analoga a quella relativa alla textbox dei termini minimi
        private void dontcare_TBox_Leave(object sender, EventArgs e)
        {
            Errore_status.Hide();

            for (int i = 0; i < dontcare_TBox.Text.Length - 1; i++)
            {
                if (dontcare_TBox.Text[i] == ',' && dontcare_TBox.Text[i + 1] == ',')
                {
                    dontcare_TBox.Text = dontcare_TBox.Text.Remove(i, 1);
                    i--;
                }
            }

            if (dontcare_TBox.Text.StartsWith(","))
            {
                dontcare_TBox.Text = dontcare_TBox.Text.Remove(0, 1);
            }

            if (dontcare_TBox.Text != "")
            {
                dontcare_s = dontcare_TBox.Text.Split(',').ToList<string>();
                bool doppioni = false;
                bool rimozione_effettuata = false;
                string nuovo_testo = "";

                dontcare_s.Sort();
                int indice = 0;
                while (indice < dontcare_s.Count - 1)
                {
                    if (dontcare_s[indice] == dontcare_s[indice + 1])
                    {
                        dontcare_s.RemoveAt(indice);
                        doppioni = true;
                    }
                    else
                        indice++;
                }

                if (doppioni == true)
                {
                    foreach (string s in dontcare_s)
                    {
                        nuovo_testo += s;
                        nuovo_testo += ",";
                    }
                    nuovo_testo = nuovo_testo.Remove(nuovo_testo.Length - 1, 1);
                    MessageBox.Show(this, "Sono stati rimossi i termini minimi doppioni inseriti.");
                    dontcare_TBox.Text = nuovo_testo;
                }

                doppioni = false;

                for (indice = 0; indice < termini_minimi_s.Count; indice++)
                {
                    foreach (string s in dontcare_s)
                    {
                        if (termini_minimi_s[indice] == s)
                        {
                            termini_minimi_s.Remove(termini_minimi_s[indice]);
                            doppioni = true;
                        }
                    }
                }


                if (doppioni == true)
                {
                    nuovo_testo = "";
                    foreach (string s in termini_minimi_s)
                    {
                        nuovo_testo += s;
                        nuovo_testo += ",";
                    }
                    nuovo_testo = nuovo_testo.Remove(nuovo_testo.Length - 1, 1);
                    MessageBox.Show(this, "Sono stati rimossi i termini minimi ritrovati anche tra le condizioni di indifferenza.");
                    termini_minimi_TBox.Text = nuovo_testo;
                }


                int n, k = 0;
                int numero_elementi = dontcare_s.Count;
                for (indice = 0; indice < dontcare_s.Count; indice++)
                {
                    if (Int32.TryParse(dontcare_s[indice], out n))
                    {
                        if (n > (int)Math.Pow(2, numero_variabili) - 1)
                        {
                            dontcare_s.Remove(dontcare_s[indice]);
                            indice--;
                            rimozione_effettuata = true;
                            k++;
                        }
                    }
                }

                if (rimozione_effettuata == true && k == numero_elementi)
                {
                    dontcare_TBox.Text = "";
                    MessageBox.Show(this, "Sono state effettuate delle rimozioni di termini minimi non consentiti. Controllare il numero di variabili", "Fuori Range", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                }

                else
                {

                    if (rimozione_effettuata == true)
                    {
                        MessageBox.Show(this, "Sono state effettuate delle rimozioni di termini minimi non consentiti. Controllare il numero di variabili", "Fuori Range", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                    }

                    nuovo_testo = "";
                    Confronta_Termini ct = new Confronta_Termini();
                    dontcare_s.Sort(0, dontcare_s.Count, ct);


                    foreach (string s in dontcare_s)
                    {
                        nuovo_testo += s;
                        nuovo_testo += ",";
                    }
                    nuovo_testo = nuovo_testo.Remove(nuovo_testo.Length - 1, 1);
                    dontcare_TBox.Text = nuovo_testo;
                }
            }
        }

        //Function: azzera_button_Click
        //Azzera le componenti del Form
        private void azzera_button_Click(object sender, EventArgs e) //Il bottone di azzeramento
        {
            termini_minimi_TBox.Text = ""; //Cancella il testo nelle textbox
            dontcare_TBox.Text = "";
            termini_minimi_TBox.Enabled = true; //Riabilita i controlli
            dontcare_TBox.Enabled = true;
            n_var_cBox.Enabled = true;
            primo_step_button.Enabled = true;
            step_successivo_button.Enabled = false;
            ins_tm_eff = false;
            output.Clear(); //Cancella l'output
            dontcare_s = new List<string>(); //Reinizializza le variabili
            termini_minimi_s = new List<string>();
            qm = new QuineMcClusky();
            azzera_button.Enabled = false; //Disabilita il bottone stesso
        }

        //Function: alg_min_button_Click
        //Mostra la form per la Minimizzazione da Equazione Algebrica
        private void alg_min_button_Click(object sender, EventArgs e)
        {
            Minimizzazione_da_Equazione_Algebrica alg = new Minimizzazione_da_Equazione_Algebrica();
            alg.Show(this);
        }

        //Function: primo_step_button_Click
        //Esegue il primo step di minimizzazione
        private void primo_step_button_Click(object sender, EventArgs e) //Al click del bottone Primo Step
        {
            if (ins_tm_eff == true) //Se sono stati inseriti dei termini minimi validi
            {
                //Disabilita i controlli
                termini_minimi_TBox.Enabled = false;
                dontcare_TBox.Enabled = false;
                n_var_cBox.Enabled = false;
                int n, j = 0;

                if (dontcare_TBox.Text != "") //Se sono stati inserite delle condizioni di indifferenza
                {
                    //Passa dalla lista di condizioni di indifferenza in forma testuale all'array di interi
                    
                    dontcare_s = dontcare_TBox.Text.Split(',').ToList<string>();
                    dontcare_s.Sort();
                    dontcare = new int[dontcare_s.Count];

                    foreach (string s in dontcare_s)
                    {
                        if (s != "")
                        {
                            if (Int32.TryParse(s, out n))
                                dontcare[j] = n;
                            j++;
                        }
                    }
                }

                else
                {
                    dontcare = new int[0]; //Se non ci sono condizioni di indifferenza abbiamo semplicemente
                    //un array vuoto di interi
                }

                //Viene fatta la stessa cosa coi termini minimi

                termini_minimi = new int[termini_minimi_s.Count];
                j = 0;

                foreach (string s in termini_minimi_s)
                {
                    if (Int32.TryParse(s, out n))
                        termini_minimi[j] = n;
                    j++;
                }

                //A questo punto tutti i parametri vengono passati alla classe ed è possibile procedere con l'algoritmo

                qm.numero_variabili = numero_variabili;
                qm.termini_minimi = termini_minimi;
                qm.dontcare = dontcare;

                if (fine_checkbox.Checked == true) //Se l'utente vuole direttamente arrivare alla forma minima
                {
                    qm.Primo_Step(); //Viene eseguito il primo step
                    while (qm.fine != true)
                        qm.Step_Successivo(); //e i successivi fino a quando non è più possibile semplificare

                    output.Text += "\nForma minima della funzione ottenuta tramite l'algoritmo di Quine Mc-Cluskey:\n";

                    foreach (string s in qm.Stampa_Stato_Finale()) //Viene direttamente stampato lo stato finale
                    {
                        output.Text += s;
                    }

                    output.Text += "\n\n";

                    if (petrick_checkbox.Checked == true)
                    {
                        foreach (string s in qm.Stampa_Stato_Finale_Petrick())
                        {
                            output.Text += s;
                        }
                    }

                    output.Focus();
                    output.Select(output.TextLength, 0);
                    primo_step_button.Enabled = false;
                    azzera_button.Enabled = true;
                    petrick_checkbox.Checked = false;
                    fine_checkbox.Checked = false;
                }
                else //Altrimenti
                {

                    foreach (string s in qm.Primo_Step()) //Viene soltanto stampato lo stato iniziale
                    {
                        output.Text += s;
                    }

                    output.Text += "\n";
                    output.Focus();
                    output.Select(output.TextLength, 0);


                    step_successivo_button.Enabled = true; //E viene abilitato il bottone per effettuare lo step successivo
                    primo_step_button.Enabled = false;
                }
            }
            else
            {
                MessageBox.Show(this, "Occorre inserire dei termini minimi e delle condizioni di indifferenza valide!", "Errore", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
        }
    }
}
