﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Text.RegularExpressions;
using Quine_McCluskey;

namespace Algoritmo_Quine_McCluskey
{
    public partial class Minimizzazione_da_Equazione_Algebrica : Form
    {
        //Property: sop_reg
        //Filtro per l'input in forma sop (sono ammesse solo le lettere, il + e il segno di negazione)
        public Regex sop_reg = new Regex("[^A-Z+~]");
        //Property: pos_reg
        //Filtro per l'input in forma pos (sono ammesse solo le lettere, il +, il segno di negazione, e le parentesi)
        public Regex pos_reg = new Regex("[^A-Z+()~]"); 
        //Variable: quine_mccluskey
        //La procedura si avvale dell'algoritmo Quine Mc-Cluskey e del metodo di Petrick
        public QuineMcClusky quine_mccluskey;
        //Variable: numero_variabili
        //Numero di variabili nella funzione
        public int numero_variabili = 2;
        public bool skip = false;
        public Minimizzazione_da_Equazione_Algebrica()
        {
            InitializeComponent();
        }

        
        private void Minimizzazione_da_Equazione_Algebrica_Load(object sender, EventArgs e)
        {
            sop_cbox.Checked = true; //Selezione di default la forma SOP per l'equazione algebrica in input
            sopdc_cbox.Checked = true;
        }

        private void sop_cbox_CheckedChanged(object sender, EventArgs e) //Chiaramente può essere selezionata solo una forma in ingresso tra SOP e POS
        {
            if (sop_cbox.Checked == true)
            {
                pos_cbox.Checked = false;
                equazione_tbox.Text = sop_reg.Replace(equazione_tbox.Text, ""); //Rimuove i termini non consentiti per la forma SOP
            }
            if(sop_cbox.Checked == false && pos_cbox.Checked == false)
            {
                sop_cbox.Checked = true;
                equazione_tbox.Text = sop_reg.Replace(equazione_tbox.Text, ""); 
            }
        }

        private void pos_cbox_CheckedChanged(object sender, EventArgs e) //Stessa cosa di sopra per la forma POS
        {
            if (pos_cbox.Checked == true)
            {
                sop_cbox.Checked = false;
                equazione_tbox.Text = pos_reg.Replace(equazione_tbox.Text, "");
            }
            if (sop_cbox.Checked == false && pos_cbox.Checked == false)
            {
                pos_cbox.Checked = true;
                equazione_tbox.Text = pos_reg.Replace(equazione_tbox.Text, "");
            }
        }

        private void equazione_tbox_TextChanged(object sender, EventArgs e) //Ogni qualvolta viene digitato del testo nella textbox relativa all'equazione vengono utilizzati
        //I filtri Regex
        {
            if (sop_cbox.Checked)
                equazione_tbox.Text = sop_reg.Replace(equazione_tbox.Text, "");
            if (pos_cbox.Checked)
                equazione_tbox.Text = pos_reg.Replace(equazione_tbox.Text, "");
            equazione_tbox.Focus();
            equazione_tbox.Select(equazione_tbox.Text.Length, 0);
        }

        private void minimizza_button_Click(object sender, EventArgs e)
        {
            List<string> funzione = new List<string>(), condizioni_di_indifferenza = new List<string>();
            bool t_m, c_d_i;
            equazione_tbox.Text = Aggiusta_NOT(equazione_tbox.Text);
            dc_tbox.Text = Aggiusta_NOT(dc_tbox.Text); 

            if (!String.IsNullOrEmpty(equazione_tbox.Text))
            {
                if (sop_cbox.Checked)
                    t_m = ASFD_SOP(equazione_tbox.Text, funzione);
                else
                    t_m = ASFD_POS(equazione_tbox.Text, funzione); //Aggiusta il numero di connettivi NOT
                if (!t_m)
                    MessageBox.Show("La funzione immessa non è espressa in forma corretta!");
                else
                {
                    numero_variabili = Ottieni_Numero_Variabili(equazione_tbox.Text, dc_tbox.Text);
                    quine_mccluskey = new QuineMcClusky();
                    quine_mccluskey.numero_variabili = numero_variabili;

                    if (sop_cbox.Checked)
                    {
                        Sistema_Espressioni(false, funzione);
                        equazione_tbox.Text = "";
                        foreach (string s in funzione)
                        {
                            equazione_tbox.Text += s + "+";
                        }
                        equazione_tbox.Text = equazione_tbox.Text.Remove(equazione_tbox.Text.Length - 1, 1);
                        quine_mccluskey.termini_minimi = SOP(funzione, numero_variabili).ToArray();
                    }
                    else
                    {
                        Sistema_Espressioni(true, funzione);
                        equazione_tbox.Text = "";
                        foreach (string s in funzione)
                        {
                            equazione_tbox.Text += "(" + s + ")";
                        }
                        quine_mccluskey.termini_minimi = POS(funzione, numero_variabili).ToArray();
                    }

                    if (!String.IsNullOrEmpty(dc_tbox.Text))
                    {
                        if (sopdc_cbox.Checked)
                            c_d_i = ASFD_SOP(dc_tbox.Text, condizioni_di_indifferenza);
                        else
                            c_d_i = ASFD_POS(dc_tbox.Text, condizioni_di_indifferenza);

                        if (!c_d_i)
                        {
                            MessageBox.Show("Le condizioni di indifferenza non sono espresse in forma corretta, pertanto non verranno tenute in considerazione!");
                            List<int> vuota = new List<int>();
                            quine_mccluskey.dontcare = vuota.ToArray();
                        }
                        else
                        {
                            if (sopdc_cbox.Checked)
                            {
                                Sistema_Espressioni(false, condizioni_di_indifferenza);
                                dc_tbox.Text = "";
                                foreach (string s in condizioni_di_indifferenza)
                                {
                                    dc_tbox.Text += s + "+";
                                }
                                dc_tbox.Text = dc_tbox.Text.Remove(dc_tbox.Text.Length - 1, 1);
                                quine_mccluskey.dontcare = SOP(condizioni_di_indifferenza, numero_variabili).ToArray();
                            }
                            else
                            {
                                Sistema_Espressioni(true, condizioni_di_indifferenza);
                                dc_tbox.Text = "";
                                foreach (string s in condizioni_di_indifferenza)
                                {
                                    dc_tbox.Text += "(" + s + ")";
                                }
                                quine_mccluskey.dontcare = POS(condizioni_di_indifferenza, numero_variabili).ToArray();
                            }
                        }
                    }
                    else
                    {
                        List<int> vuota = new List<int>();
                        quine_mccluskey.dontcare = vuota.ToArray();
                    }
                    if (quine_mccluskey.termini_minimi.Count() == 0)
                    {
                        output_tbox.Text = "f = 0";
                    }
                    else
                    {
                        output_tbox.Text += "Termini minimi della funzione: ";
                        foreach(int n in quine_mccluskey.termini_minimi)
                        {
                            output_tbox.Text += n.ToString() + ",";
                        }
                        output_tbox.Text = output_tbox.Text.Remove(output_tbox.TextLength - 1, 1) + "\n";
                        quine_mccluskey.Primo_Step();
                        while (quine_mccluskey.fine != true)
                            quine_mccluskey.Step_Successivo();
                        foreach (string s in quine_mccluskey.Stampa_Stato_Finale_Petrick())
                        {
                            output_tbox.Text += s;
                        }
                    }
                }
            }
            else
            {
                MessageBox.Show("Inserire un input valido e non nullo!");
            }
        }

        private void equazione_tbox_Leave(object sender, EventArgs e)
        {
        }

        private void azzera_button_Click(object sender, EventArgs e) // Bottone di azzeramento, come in Main.cs
        {
            output_tbox.Text = "";
            equazione_tbox.Text = "";
            dc_tbox.Text = "";
            equazione_tbox.Focus();
            equazione_tbox.Select(equazione_tbox.Text.Length, 0);
            skip = false;
        }

        private void copia_toolstrip_Click(object sender, EventArgs e)
        {
            if (output_tbox.SelectedText != "")
                Clipboard.SetText(output_tbox.SelectedText);
        }

        private void Minimizzazione_da_Equazione_Algebrica_HelpButtonClicked(object sender, CancelEventArgs e)
        {
            MessageBox.Show(this,
                "Tramite questa funzionalità è possibile semplificare una funzione logica inserita direttamente nella sua forma algebrica SOP o POS\n" +
                "Un esempio di funzione logica in forma SOP è il seguente: f = ~DEF + ~ACF + ACE + ~A~B~D~E~F + ~AB~CD + A~B~C\n" +
                "Il simbolo ~ rappresenta l'operatore logico di negazione.\n" + 
                "Un esempio di funzione logica in forma POS è il seguente f = (~A + B)(~C + ~B + D)(~D + E)\n" +
                "Inoltre è possibile inserire le condizioni di indifferenza come termine booleano in forma SOP o POS.\n" +
                "Verranno utilizzati l'algoritmo di Quine Mc-Cluskey e il metodo di Petrick per la semplificazione."
                , "Aiuto", MessageBoxButtons.OK, MessageBoxIcon.Question, MessageBoxDefaultButton.Button1);
        }

        private void posdc_cbox_CheckedChanged(object sender, EventArgs e)
        {
            if (posdc_cbox.Checked == true)
            {
                sopdc_cbox.Checked = false;
                dc_tbox.Text = pos_reg.Replace(dc_tbox.Text, "");
            }
            if (sopdc_cbox.Checked == false && posdc_cbox.Checked == false)
            {
                posdc_cbox.Checked = true;
                dc_tbox.Text = pos_reg.Replace(dc_tbox.Text, "");
            }
        }

        private void sopdc_cbox_CheckedChanged(object sender, EventArgs e)
        {
            if (sopdc_cbox.Checked == true)
            {
                posdc_cbox.Checked = false;
                dc_tbox.Text = sop_reg.Replace(dc_tbox.Text, ""); //Rimuove i termini non consentiti per la forma SOP
            }
            if (sopdc_cbox.Checked == false && posdc_cbox.Checked == false)
            {
                sopdc_cbox.Checked = true;
                dc_tbox.Text = sop_reg.Replace(dc_tbox.Text, "");
            }
        }

        private void dc_tbox_TextChanged(object sender, EventArgs e)
        {
            if (sopdc_cbox.Checked)
                dc_tbox.Text = sop_reg.Replace(dc_tbox.Text, "");
            if (posdc_cbox.Checked)
                dc_tbox.Text = pos_reg.Replace(dc_tbox.Text, "");
            dc_tbox.Focus();
            dc_tbox.Select(dc_tbox.Text.Length, 0);
        }

        private void dc_tbox_Leave(object sender, EventArgs e)
        {
        }

        //Function: Aggiusta_NOT
        //Metodo che elimina i connettivi NOT superflui
        private string Aggiusta_NOT(string espr) 
        {
            string espressione = espr;
            int n = 0, j = 0; ;

            for (int i = 0; i < espressione.Length; i++)
            {
                if (espressione[i] == '~')
                {
                    j = i;
                    while (espressione[j] == '~')
                    {
                        n++;
                        j++;
                    }
                    if (n % 2 == 0)
                    {
                        espressione = espressione.Remove(i, j - i);
                    }
                    else
                    {
                        espressione = espressione.Remove(i, j - i - 1);
                    }
                    n = 0;
                }
            }
            return espressione;
        }

        /*Function: ASFD_POS
        Parser per la forma POS basato sull'idea di automa a stati finiti deterministico. L'automa in questo caso,
        lavora con sei stati e un alfabeto di 30 simboli {'(', ')', '+', '~', 'A', 'B', ... , 'Z'}.

        Returns:
        false se la stringa non appartiene al linguaggio, true viceversa.
        */

        private bool ASFD_POS(string espr, List<string> terms)
        {
            int state = 0;
            string term = "";
            foreach (char c in espr)
            {
                if (c != '(' && c != ')')
                    term = term + c;
                switch (state)  //Funzione di transizione
                {
                    case 0:
                        if (c == '~')
                        {
                            state = 4;
                        }
                        else if (c == '+')
                        {
                            state = 6;
                        }
                        else if (c >= 65 && c <= 90)
                        {
                            state = 3;
                        }
                        else if (c == '(')
                        {
                            state = 1;
                        }
                        else if (c == ')')
                        {
                            state = 6;
                        }
                        else
                        {
                            state = 6;
                        }
                        break;
                    case 1:
                        if (c == '~')
                        {
                            state = 5;
                        }
                        else if (c == '+')
                        {
                            state = 5;
                        }
                        else if (c >= 65 && c <= 90)
                        {
                            state = 2;
                        }
                        else if (c == '(')
                        {
                            state = 6;
                        }
                        else if (c == ')')
                        {
                            state = 6;
                        }
                        else
                        {
                            state = 6;
                        }
                        break;
                    case 2:
                        if (c == '~')
                        {
                            state = 6;
                        }
                        else if (c == '+')
                        {
                            state = 1;
                        }
                        else if (c >= 65 && c <= 90)
                        {
                            state = 5;
                        }
                        else if (c == '(')
                        {
                            state = 5;
                        }
                        else if (c == ')')
                        {
                            state = 0;
                        }
                        else
                        {
                            state = 6;
                        }
                        break;
                    case 3:
                        if (c == '~')
                        {
                            state = 4;
                        }
                        else if (c == '+')
                        {
                            state = 6;
                        }
                        else if (c >= 65 && c <= 90)
                        {
                            state = 3;
                        }
                        else if (c == '(')
                        {
                            state = 1;
                        }
                        else if (c == ')')
                        {
                            state = 6;
                        }
                        else
                        {
                            state = 6;
                        }
                        break;

                    case 4:
                        if (c == '~')
                        {
                            state = 6;
                        }
                        else if (c == '+')
                        {
                            state = 6;
                        }
                        else if (c >= 65 && c <= 90)
                        {
                            state = 3;
                        }
                        else if (c == '(')
                        {
                            state = 6;
                        }
                        else if (c == ')')
                        {
                            state = 6;
                        }
                        else
                        {
                            state = 6;
                        }
                        break;

                    case 5:
                        if (c == '~')
                        {
                            state = 6;
                        }
                        else if (c == '+')
                        {
                            state = 6;
                        }
                        else if (c >= 65 && c <= 90)
                        {
                            state = 2;
                        }
                        else if (c == '(')
                        {
                            state = 6;
                        }
                        else if (c == ')')
                        {
                            state = 6;
                        }
                        else
                        {
                            state = 6;
                        }
                        break;
                    case 6:
                        terms.Clear();
                        return false;
                }
                if (state == 0 || state == 3) //Lo stato di accettazione è lo stato 0.
                {
                    terms.Add(term);
                    term = "";
                }
            }
            if (state == 0 || state == 3)
            {
                if (!String.IsNullOrEmpty(term))
                    terms.Add(term);
                return true;
            }

            terms.Clear();
            return false;
        }

        /*Function: ASFD_SOP
        Parser per la forma SOP basato sull'idea di automa a stati finiti deterministico. L'automa in questo caso,
        lavora con qattro stati e un alfabeto di 28 simboli {'+', '~', 'A', 'B', ... , 'Z'}.

        Returns:
        false se la stringa non appartiene al linguaggio, true viceversa.
        */
        private bool ASFD_SOP(string espr, List<string> terms)
        {
            int state = 0;
            string term = "";
            foreach (char c in espr)
            {
                if (c != '+')
                    term = term + c;
                switch (state)
                {
                    case 0:
                        if (c == '~')
                        {
                            state = 1;
                        }
                        else if (c == '+')
                        {
                            state = 2;
                        }
                        else if (c >= 65 && c <= 90)
                        {
                            state = 0;
                        }
                        else
                        {
                            state = 3;
                        }
                        break;
                    case 1:
                        if (c == '~')
                        {
                            state = 3;
                        }
                        else if (c == '+')
                        {
                            state = 3;
                        }
                        else if (c >= 65 && c <= 90)
                        {
                            state = 0;
                        }
                        else
                        {
                            state = 3;
                        }
                        break;
                    case 2:
                        if (c == '~')
                        {
                            state = 1;
                        }
                        else if (c == '+')
                        {
                            state = 3;
                        }
                        else if (c >= 65 && c <= 90)
                        {
                            state = 0;
                        }
                        else
                        {
                            state = 3;
                        }
                        break;
                    case 3:
                        terms.Clear();
                        return false;
                }
                if (state == 2) //Separa i vari termini in base al simbolo '+'
                {
                    terms.Add(term);
                    term = "";
                }
            }
            if (state == 0)
            {
                terms.Add(term);
                return true;
            }

            terms.Clear();
            return false;
        }

        //Function: Ottieni_Numero_Variabili
        //Ricavera il numero di variabili della funzione dall'input dell'utente
        private int Ottieni_Numero_Variabili(string espressione, string condizioni_indifferenza) //Ricava il numero di variabili dall'input dell'utente
        {
            int n_var = 0;
            foreach(char c in espressione)
            {
                if(c >= 65 && c <= 90)
                {
                    if (n_var < (c - 64))
                        n_var = c - 64;
                }
            }
            foreach (char c in condizioni_indifferenza)
            {
                if (c >= 65 && c <= 90)
                {
                    if (n_var < (c - 64))
                        n_var = c - 64;
                }
            }
            return n_var;
        }

        //Function: SOP
        //Ricava la tabella di veritè della funzione in forma SOP
        private List<int> SOP(List<string> sotto_espressioni, int n_var)
        {
            List<int> termini = new List<int>();

            bool gruppo = false, addendo = true;
            string rappresentazione_binaria;

            for (int i = 0; i < (int)(Math.Pow(2, n_var)); i++)
            {
                rappresentazione_binaria = Converti_In_Binario(i, n_var);
                foreach(string sub_espr in sotto_espressioni)
                {
                    for (int j = 0; j < sub_espr.Length; j++)
                    {
                        if (sub_espr[j] == '~')
                            continue;

                        if (j > 0)
                        {
                            if (rappresentazione_binaria[(int)(sub_espr[j] - 65)] == '0' && sub_espr[j - 1] == '~')
                            {
                                addendo &= true;
                                continue;
                            }
                            if (rappresentazione_binaria[(int)(sub_espr[j] - 65)] == '1' && sub_espr[j - 1] == '~')
                            {
                                addendo &= false;
                                break;
                            }
                            if (rappresentazione_binaria[(int)(sub_espr[j] - 65)] == '0')
                            {
                                addendo &= false;
                                break;
                            }
                            if (rappresentazione_binaria[(int)(sub_espr[j] - 65)] == '1')
                            {
                                addendo &= true;
                                continue;
                            }
                        }
                        else
                        {

                            if (rappresentazione_binaria[(int)(sub_espr[j] - 65)] == '0')
                            {
                                addendo &= false;
                                break;
                            }
                            if (rappresentazione_binaria[(int)(sub_espr[j] - 65)] == '1')
                            {
                                addendo &= true;
                                continue;
                            }
                        }
                    }
                    gruppo |= addendo;
                    addendo = true;

                    if (gruppo == true)
                        break;
                }

                if (gruppo == true)
                    termini.Add(i);
                gruppo = false;
                }
            return termini;
        }

        //Function: POS
        //Ricava la tabella di veritè della funzione in forma POS
        private List<int> POS(List<string> sotto_espressioni, int n_var)
        {
            List<int> termini = new List<int>();

            bool gruppo = true, moltiplicando = false;
            string rappresentazione_binaria;

            for (int i = 0; i < (int)(Math.Pow(2, numero_variabili)); i++)
            {
                rappresentazione_binaria = Converti_In_Binario(i, n_var);
                foreach(string sub_espr in sotto_espressioni)
                {
                    for (int j = 0; j < sub_espr.Length; j++)
                    {
                        if(sub_espr[j] == '+')
                            continue;

                        if (sub_espr[j] == '~')
                            continue;

                        if (j > 0)
                        {
                            if (rappresentazione_binaria[(int)(sub_espr[j] - 65)] == '0' && sub_espr[j - 1] == '~')
                            {
                                moltiplicando |= true;
                            }
                            else if (rappresentazione_binaria[(int)(sub_espr[j] - 65)] == '1' && sub_espr[j - 1] == '~')
                            {
                                moltiplicando |= false;
                            }
                            else if (rappresentazione_binaria[(int)(sub_espr[j] - 65)] == '0')
                            {
                                moltiplicando |= false;
                            }
                            else if (rappresentazione_binaria[(int)(sub_espr[j] - 65)] == '1')
                            {
                                moltiplicando |= true;
                            }
                        }
                        else
                        {

                            if (rappresentazione_binaria[(int)(sub_espr[j] - 65)] == '0')
                            {
                                moltiplicando |= false;
                            }
                            else if (rappresentazione_binaria[(int)(sub_espr[j] - 65)] == '1')
                            {
                                moltiplicando |= true;
                            }
                        }
                    }
                    gruppo &= moltiplicando;
                    moltiplicando = false;

                    if (gruppo == false)
                        break;
                }

                if (gruppo == true)
                    termini.Add(i);
                gruppo = true;
            }
            return termini;
        }

        //Function: Converti_In_Binario
        //Procedura per la conversione in binario
        private static string Converti_In_Binario(int n, int numero_variabili)
        {
            string output = "";

            while (n >= 1)
            {
                output += (n % 2 == 1 ? "1" : "0");
                n /= 2;
            }

            while (output.Length < numero_variabili)
                output += "0";

            output = Reverse(output);
            return output;
        }

        //Function: Reverse
        //Procedura per l'inversione di una stringa
        private static string Reverse(string s)
        {
            StringBuilder sb = new StringBuilder();
            for (int i = s.Length - 1; i >= 0; i--)
            {
                sb.Append(s[i]);
            }
            return sb.ToString();
        }

        //Function: Sistema_Espressioni
        //Riformatta l'espressione inserita dall'utente, eliminando le ripetizioni non necessarie e riordinandone gli elementi.
        private void Sistema_Espressioni(bool pos, List<string> termini)
        {
            Rimuovi_Duplicati(termini);

            List<string> riarrangiata = new List<string>();
            string termine = "";

            if (pos)
            {
                List<string> addendi = new List<string>();
                foreach (string t in termini)
                {
                    foreach (string a in t.Split(new[] { '+' }, StringSplitOptions.RemoveEmptyEntries))
                    {
                        if (!addendi.Contains(a))
                            addendi.Add(a);
                    }
                    foreach (string a in addendi)
                    {
                        termine += a + "+";
                    }
                    termine = termine.Remove(termine.Length - 1, 1);
                    riarrangiata.Add(termine);
                    termine = "";
                    addendi.Clear();
                }
                termini.Clear();
            }
            else
            {
                List<string> moltiplicandi = new List<string>();
                foreach (string t in termini)
                {
                    string[] singoli_caratteri = Regex.Split(t, "");
                    for (int i = 0; i < singoli_caratteri.Length; i++)
                    {
                        if (singoli_caratteri[i] == "~")
                        {
                            moltiplicandi.Add(singoli_caratteri[i] + singoli_caratteri[i + 1]);
                            i++;
                        }
                        else
                        {
                            moltiplicandi.Add(singoli_caratteri[i]);
                        }
                    }
                    Rimuovi_Duplicati(moltiplicandi);
                    foreach (string m in moltiplicandi)
                    {
                        termine += m;
                    }
                    riarrangiata.Add(termine);
                    termine = "";
                    moltiplicandi.Clear();
                }
                termini.Clear();
            }

            Rimuovi_Duplicati(riarrangiata);
            foreach (string t in riarrangiata)
                termini.Add(t);
        }

        //Function: Rimuovi_Duplicati
        //Rimuove i termini duplicati da una lista
        private void Rimuovi_Duplicati(List<string> lista)
        {
            lista.Sort((x, y) => string.Compare(x.Replace("~", ""), y.Replace("~", "")));
            int indice = 0;
            while (indice < lista.Count - 1)
            {
                if (lista[indice] == lista[indice + 1])
                    lista.RemoveAt(indice);
                else
                    indice++;
            }
        }
    }
}
