﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Text.RegularExpressions;
using Ordinamento;

namespace Quine_McCluskey
{
    /*Class: QuineMcCluskey
    Classe che gestisce la tabella dei termini minimi e la riduzione della funzione logica corrisponente
    */
    public class QuineMcClusky
    {
        //Property: termini_minimi
        //I termini minimi della funzione
        public int[] termini_minimi { get; set; }
        //Property: dontcare
        //Le condizioni di indifferenza
        public int[] dontcare { get; set; }
        //Property: numero_variabili
        //Il numero di variabili logiche nella funzione. Il massimo numero di variabili è 26(rappresentate dalle lettere maiuscole dalla A alla Z)
        public int numero_variabili; //Il numero di variabili dell'espressione
        //Variable: rappresentazioni_binarie
        //Lista di stringhe contenente le rappresentazioni binarie degli implicanti.
        private List<string> rappresentazioni_binarie = new List<string>();
        //Variable: temp
        //Una lista temporanea di stringhe utilizzata durante la computazione. Conterrà semplicemente una copia della lista precedente.
        private List<string> temp = new List<string>();
        //Variable: termini_essenziali
        //La lista dei termini essenziali deve restare separata da quella delle rappresentazioni binarie perchè può contenere dei termini che durante lo svolgimento
        //dell'algoritmo non vanno incontro a semplificazione e rimangono isolati.
        private List<string> termini_essenziali = new List<string>();
        //Variable: gruppi
        //Il numero di elementi per ogni gruppo, l'indice dell'array rappresenta il gruppo.
        private int[] gruppi;
        //Variable: fine
        //Indica che l'algoritmo di semplificazione è giunto al termine.
        //(La ricerca degli implicanti primi essenziali viene effettuata dopo che l'algoritmo di semplificazione è terminato).
        public bool fine;


        /*Function: Combinazioni
        Permette di risalire ai termini minimi implicati nella riduzione partendo da una rappresentazione binaria ridotta

        Returns:
        Tutte le possibili combinazioni ottenute sostituendo agli underscores nella stringa o 0 o 1.*/

        private IEnumerable<string> Combinazioni(string input)
        {
            var testa = input[0] == '_'
                ? new[] { "0", "1" }
                : new[] { input[0].ToString() };

            var coda = input.Length > 1
                ? Combinazioni(input.Substring(1))
                : new[] { "" };

            return
                from t in testa
                from c in coda
                select t + c;
        }

        /*Function: Converti_in_Binario
        Questa funzione converte in binario tutti gli elementi degli array numerici termini_minimi e dontcare e aggiunge ogni
        rappresentazione binaria ottenuta alla lista rappresentazioni_binarie.*/
        private void Converti_in_Binario()
        {
            char[] binario = new char[numero_variabili];

            int i, j;
            foreach (int t_m in termini_minimi)
            {
                j = t_m;
                
                for (i = 0; i < numero_variabili; i++) //La stringa viene inizializzata con degli zero.
                {
                    binario[i] = '0';
                }
                i = numero_variabili - 1;

                while (j != 0)
                {
                    if (j % 2 == 0)
                    {
                        binario[i] = '0';
                    }
                    else
                    {
                        binario[i] = '1';
                    }
                    j /= 2;
                    i--;
                }
                string r_b = new string(binario); //Conversione dell'array di caratteri in stringa

                rappresentazioni_binarie.Add(r_b); //Aggiunge la stringa ottenuta alla lista rappresentazioni_binarie.
            }

            foreach (int dc in dontcare) //Stessa cosa per le condizioni di indifferenza
            {
                j = dc;

                for (i = 0; i < numero_variabili; i++)
                {
                    binario[i] = '0';
                }
                i = numero_variabili - 1;

                while (j != 0)
                {
                    if (j % 2 == 0)
                    {
                        binario[i] = '0';
                    }
                    else
                    {
                        binario[i] = '1';
                    }
                    j /= 2;
                    i--;
                }
                string r_b = new string(binario);
                rappresentazioni_binarie.Add(r_b);
            }

            //Eventualmente aggiungere!
            //Rimuovi_Duplicati(rappresentazioni_binarie);
            //Questa funzione non è necessaria in questa specifica versione del programma poichè viene fatto un controllo
            //a monte sull'input dell'utente. Ho pensato però che potesse essere utile mantenerla nella classe.

            Ordina_per_Gruppi(); //Chiama la funzione interna di ordinamento per gruppi
        }

        /*Function: Converti_In_Decimale
        Converte in decimale una rappresentazione binaria */
        private int Converti_In_Decimale(string n_bin)
        {
            StringBuilder sb = new StringBuilder();
            int i;
            for (i = n_bin.Length - 1; i >= 0; i--)
            {
                sb.Append(n_bin[i]);
            }

            string bin = sb.ToString();
            int dec = 0;

            for (i = 0; i < bin.Length; i++)
            {
                dec += (int)Math.Pow(2, i) * (bin[i] - 48); //Sottrendo 48 a un carattere se questo rappresentava un numero
                //otteniamo il numero rappresentato
            }
            return dec;
        }

        //Function: Conta_Gruppo
        //Funzione che conta gli uno presenti nella rappresentazione e ne restituisce il gruppo di appartenenza
        private int Conta_Gruppo(string rappresentazione_binaria) 
        {
            //Code:
            int gruppo = rappresentazione_binaria.Split('1').Length - 1; //Viene fatto il conteggio delle stringhe risultanti
            //dalla separazione in sottostringhe con carattere di separazione '1'.

            return gruppo;
        }

        //Function: Ordina_per_Gruppi
        //Funzione che ordina la lista rappresentazioni_binarie in funzione dei gruppi
        private void Ordina_per_Gruppi() //Funzione che ordina la lista rappresentazioni_binarie in funzione dei gruppi
        {
            string temp;
            int gruppo_1, gruppo_2; //Viene fatta la comparazione tra il gruppo di una stringa nella lista e quello della
            //successiva

            int i, j;

            for (i = rappresentazioni_binarie.Count - 1; i >= 0; i--)
            {
                for (j = 1; j <= i; j++)
                {
                    gruppo_1 = Conta_Gruppo(rappresentazioni_binarie[j - 1]);
                    gruppo_2 = Conta_Gruppo(rappresentazioni_binarie[j]);
                    if (gruppo_1 > gruppo_2)
                    {
                        temp = rappresentazioni_binarie[j - 1];
                        rappresentazioni_binarie[j - 1] = rappresentazioni_binarie[j];
                        rappresentazioni_binarie[j] = temp;
                    }
                }
            }
        }
        //Function: Gruppo_Massimo
        //Restituisce il gruppo massimo tra gli implicanti
        private int Gruppo_Massimo()
        {
            int g_max = 0;
            foreach (string s in rappresentazioni_binarie)
            {
                if (Conta_Gruppo(s) > g_max)
                    g_max = Conta_Gruppo(s);
            }

            return g_max;
        }
        //Function Inizializza_Gruppi
        //Inizializza l'array numerico gruppi
        private void Inizializza_Gruppi()
        {
            gruppi = new int[Gruppo_Massimo() + 1];
            int a = 0;
            for (int i = 0; i < gruppi.Length; i++)
            {
                foreach (string s in rappresentazioni_binarie)
                {
                    if (Conta_Gruppo(s) == i)
                        a++;
                }
                gruppi[i] = a;
                a = 0;
            }
        }

        /*Function: Confronta
        Funzione che confronta due stringhe e restituisce una nuova stringa qualora sia stata riscontrata solo una differenza.
        (una posizione in cui differisce il valore del singolo carattere 0 o 1)
        La nuova stringa è uguale a quelle di partenza con l'eccezione che nella posizione in cui è stata 
        individuata la differenza viene inserito un '_'. Se per esempio venissero comparate le stringhe "0001" e "1001" l'unica 
        differenza sarebbe in posizione 0 e la stringa risultate sarebbe "_001". Se invece venissero comparate le stringhe "0011" 
        e "1001" avremmo due differenze, rispettivamente in posizione 0 e 2, pertanto non verrebbe restituita alcuna stringa.*/
        private string Confronta(string a, string b)
        {
            int numero_differenze = 0;
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < a.Length; i++)
            {
                if (a[i] != b[i])
                {
                    numero_differenze++;
                    sb.Append('_');
                }
                else
                {
                    sb.Append(a[i]);
                }
            }
            if (numero_differenze == 1)
                return sb.ToString();
            return null;
        }
        //Function: Rimuovi_Duplicati
        //Rimuove i termini duplicati da una lista
        private void Rimuovi_Duplicati(List<string> lista)
        {
            lista.Sort();
            int indice = 0;
            while (indice < lista.Count - 1)
            {
                if (lista[indice] == lista[indice + 1])
                    lista.RemoveAt(indice);
                else
                    indice++;
            }
        }
        //Function: Rimuovi_Termini_non_Essenziali
        //Rimuove i termini che non sono essenziali dalla lista termini_essenziali
        //Questo è necessario poichè può avvenire che un termine in realtà non essenziale venga qui aggiunto.
        private void Rimuovi_Termini_non_Essenziali() 

        {
            for (int i = 0; i < gruppi.Length - 1; i++) //Per ogni gruppo
            {
                foreach (string a in rappresentazioni_binarie) //Vengono confrontate le rappresentazioni binarie
                {
                    if (Conta_Gruppo(a) == i) //Di gruppo i
                    {
                        foreach (string b in rappresentazioni_binarie)
                        {
                            if (Conta_Gruppo(b) == i + 1) // Con quelle di gruppo i + 1
                            {
                                if (Confronta(a, b) != null) //Se è possibile semplificarle
                                {
                                    termini_essenziali.Remove(a); //Allora non
                                    termini_essenziali.Remove(b); //sono termini essenziali
                                }
                            }
                        }
                    }
                }
            }
            for (int i = gruppi.Length - 1; i > 0; i--) //Lo stesso procedimento viene fatto dal basso verso l'alto
            {
                foreach (string a in rappresentazioni_binarie)
                {
                    if (Conta_Gruppo(a) == i)
                    {
                        foreach (string b in rappresentazioni_binarie)
                        {
                            if (Conta_Gruppo(b) == i - 1)
                            {
                                if (Confronta(a, b) != null)
                                {
                                    termini_essenziali.Remove(a);
                                    termini_essenziali.Remove(b);
                                }
                            }
                        }
                    }
                }
            }
        }

        //Function: Primo_Step
        //Questa funzione restituisce una lista stampabile contenente semplicemente
        //i termini minimi e le condizioni di indifferenza rappresentate in binario con il gruppo di appartenenza
        public List<string> Primo_Step()
        {
            Converti_in_Binario();
            return Stampa_Stato_Corrente(); //Viene sfruttata questa funzione interna che viene descritta successivamente
        }
        //Function: Step_Successivo
        //Procede di un passo avanti con la riduzione
        public void Step_Successivo()
        {
            Rimuovi_Duplicati(rappresentazioni_binarie); //Vengono eliminati dei termini duplicati che si vengono a creare
            //ad ogni step
            Inizializza_Gruppi(); //Ad ogni step chiaramente i gruppi vengono inizializzati nuovamente

            temp.Clear(); //Viene pulita la lista temporanea

            int k = 0;
            for (int i = 0; i < gruppi.Length - 1; i++) //Per ogni gruppo
            {
                foreach (string a in rappresentazioni_binarie) //Vengono confrontate le rappresentazioni binarie
                {
                    if (Conta_Gruppo(a) == i) //di gruppo i
                    {
                        foreach (string b in rappresentazioni_binarie)
                        {
                            if (Conta_Gruppo(b) == i + 1) //con quelle di gruppo i + 1.
                            {
                                if (Confronta(a, b) != null) //Se è possibile semplificare
                                {
                                    temp.Add(Confronta(a, b)); // il risultato della semplificazione viene aggiunto
                                    //alla lista temporanea.
                                }
                                else //Altrimenti
                                {
                                    k++; //Viene incrementato l'indice k, che serve a verificare che il termine analizzato
                                    //non sia essenziale.
                                    //Questo si rivela essenziale qualora non sia stato possibile semplificarlo con nessun
                                    //altro termine.
                                }
                            }
                        }
                        if (k == gruppi[i + 1]) //Se l'indice k è uguale al numero di elementi del gruppo i + 1
                        {
                            termini_essenziali.Add(a); //Viene aggiunta la rappresentazione binaria in esame
                            //ai termini essenziali
                        }
                    }
                    k = 0; //L'indice viene azzerato ad ogni passaggio
                }
            }


            for (int i = gruppi.Length - 1; i > 0; i--) //Viene effettuato nuovamente il passaggio dal basso verso l'alto
            {
                foreach (string a in rappresentazioni_binarie)
                {
                    if (Conta_Gruppo(a) == i)
                    {
                        foreach (string b in rappresentazioni_binarie)
                        {
                            if (Conta_Gruppo(b) == i - 1)
                            {
                                if (Confronta(a, b) == null)
                                {
                                    k++;
                                }
                            }
                        }
                        if (k == gruppi[i - 1])
                        {
                            termini_essenziali.Add(a);
                        }
                    }
                    k = 0;
                }
            }

            Rimuovi_Termini_non_Essenziali(); //Vengono eliminati i termini non essenziali
            Rimuovi_Duplicati(termini_essenziali); //Vengono eliminati i duplicati

            if (gruppi.Length == 1) //Se è rimasto solo un gruppo allora ogni termine è essenziale e va direttamente in termini_essenziali
            {
                foreach (string s in rappresentazioni_binarie)
                {
                    termini_essenziali.Add(s);
                }
            }

            if (temp.Count != 0) //Se la lista temp (contenente i termini trovati precedentemente tramite semplificazione) 
            //non è vuota
            {
                rappresentazioni_binarie.Clear(); //viene pulita la lista rappresentazioni_binarie
                foreach (string s in temp) //e ogni stringa contenuta in temp
                {
                    rappresentazioni_binarie.Add(s); //viene ivi trasferita.
                }
                fine = false; //In questo caso la semplificazione ancora può procedere.
            }
            else
            {
                fine = true; //Se non vi sono più termini semplificabili allora l'algoritmo deve fermarsi.
            }
        }

        //Function: Stampa_Stato_Corrente
        //Restituisce una lista stampabile dello stato corrente 
        public List<string> Stampa_Stato_Corrente()
        {
            List<string> stato_corrente = new List<string>();
            string linea = ""; //Singola linea da stampare
            List<int> t_m = new List<int>(); //Lista di stringhe dei termini minimi (ottenuti tramite la funzione Combinazioni)
            Rimuovi_Duplicati(rappresentazioni_binarie);

            foreach (string s in rappresentazioni_binarie)
            {
                linea = ""; //La linea viene inizializzata a stringa nulla e conterrà
                linea += "Gruppo: "; //Il gruppo
                linea += Conta_Gruppo(s);
                linea += " Termini Minimi: "; //I termini minimi implicati
                foreach (string c in Combinazioni(s))
                {
                    t_m.Add(Converti_In_Decimale(c));
                }
                foreach (int n in t_m)
                {
                    linea += "M" + n.ToString() + "-";
                }
                linea = linea.Remove(linea.Length - 1, 1);
                linea += " Rappresentazione Binaria: " + s; //La rappresentazione binaria in esame
                linea += "\n"; //Il carriage return
                stato_corrente.Add(linea);
                t_m.Clear();
            }

            return stato_corrente;
        }
        //Function: Implicanti_Primi_Essenziali
        //Questa funzione serve nell'ultimo step a eliminare tutti i termini non essenziali, effettuando la ricerca della "copertura ottima"
        private List<string> Implicanti_Primi_Essenziali()
        {
            List<string> Implicanti_Primi_Essenziali = new List<string>(); //Lista che conterrà i risultati dello step finale

            bool[] verificato = new bool[(int)Math.Pow(2, numero_variabili)]; //Tabella completa di ogni termine.
            //I termini in totale sono 2 elevato al numero di variabili della funzione.
            //La tabella è una tabella di bool poichè ci interessa verificare semplicemente che ogni termine minimo
            //sia coperto dagli implicanti primi essenziali.

            for (int i = 0; i < verificato.Length; i++)
            {
                verificato[i] = false; //Vengono inizializzati tutti a false.
            }

            string[] l1;
            int contatore = 0;
            bool salta = false;

            foreach (string s in temp) //Per ogni stringa in temp (che contiene i termini essenziali ricavati dalle rappresentazioni binarie finali
            //, separati dal carattere '-', ottenuti terminata l'applicazione dell'algoritmo di semplificazione e 
            //chiamata la funzione Ultimo_Passaggio()

            {
                l1 = s.Split('-'); //Separa i termini essenziali in un'array di stringhe

                foreach (string a in l1) //Per ogni termine essenziale
                {
                    foreach (int dc in dontcare) //Se questo termine è una condizione di indifferenza
                    {
                        if (a == dc.ToString())
                        {
                            salta = true; //allora salta alla prossima stringa di termini essenziali
                            break;
                        }
                    }
                    if (salta == false)
                    {
                        foreach (string b in temp) //Verifica che la combinazione di termini essenziali sia imprescindibile
                        //Ciò è tale se in nessun altra stringa di termini essenziali figurano i termini essenziali della stringa in oggetto
                        {
                            if (b != s)
                            {
                                if (E_Contenuto(b, a) == true)
                                {
                                    contatore++;
                                }
                            }
                        }
                        if (contatore == 0) //Infatti questa viene aggiunta alla lista Implicanti_Primi_Essenziali solo se
                        //non è stato trovato nessun riscontro (la variabile contatore contiene il numero di riscontri).
                        {
                            Implicanti_Primi_Essenziali.Add(s);
                            foreach (string b in l1)
                            {
                                verificato[Convert.ToInt32(b)] = true; //Viene registrato il fatto che il singolo termine
                                //è stato coperto
                            }
                        }
                        contatore = 0; //viene ripristinato il contatore
                    }

                    salta = false; //viene inizializzata nuovamente la variabile salta per il prossimo ciclo
                }
            }

            foreach (int a in termini_minimi) //per ogni termine minimo dato inizialmente in input dall'utente
            {
                if (verificato[a] == false) //se uno di questi non è ancora stato coperto
                {
                    foreach (string s in temp)
                    {
                        if (E_Contenuto(s, a.ToString()) == true) //viene presa la prima stringa di termini essenziali che
                        //lo contiene
                        {
                            Implicanti_Primi_Essenziali.Add(s); //e viene aggiunta alla lista Implicanti_Primi_Essenziali
                            l1 = s.Split('-');
                            foreach (string b in l1)
                            {
                                verificato[Convert.ToInt32(b)] = true; //ancora una volta viene registrata l'avvenuta copertura
                            }
                            break; //Il foreach viene interrotto perchè viene presa la prima stringa utile, non è importante
                            //quale sia.
                        }
                    }
                }
            }

            Rimuovi_Duplicati(Implicanti_Primi_Essenziali);

            return Implicanti_Primi_Essenziali;
        }
        //Function: E_Contenuto
        //Funzione che verifica che un termine essenziale sia contenuto in una stringa di termini essenziali
        private bool E_Contenuto(string a, string b)
        {
            string[] cont = a.Split('-');
            foreach (string min in cont)
            {
                if (min == b)
                    return true;
            }
            return false;
        }
        //Function: Ultimo_Passaggio
        //Nell'ultimo passaggio le rappresentazioni binarie vengono convertite in stringhe di termini essenziali
        private void Ultimo_Passaggio()
        {
            temp.Clear();
            string cont = "";
            foreach (string s in termini_essenziali)
            {
                foreach (string c in Combinazioni(s).ToList())
                {
                    cont += Converti_In_Decimale(c).ToString() + "-";
                }
                cont = cont.Remove(cont.Length - 1, 1);
                temp.Add(cont);
                cont = "";
            }
        }

        //Function: Stampa_Stato_Finale
        //Funzione analoga a Stampa_Stato_Corrente() con la differenza che questa volta viene restituita la forma SOP finale della funzione semplificata.
        public List<string> Stampa_Stato_Finale()
        {
            List<string> stato_finale = new List<string>();

            Ultimo_Passaggio();
            rappresentazioni_binarie = Implicanti_Primi_Essenziali();

            StringBuilder sb = new StringBuilder();
            char c;
            sb.Append("f = ");
            foreach (string s in rappresentazioni_binarie)
            {
                for (int i = 0; i < termini_essenziali[temp.IndexOf(s)].Length; i++)
                {
                    if (termini_essenziali[temp.IndexOf(s)][i] == '_')
                        sb.Append("");
                    if (termini_essenziali[temp.IndexOf(s)][i] == '0') //Se zero viene negato
                    {
                        c = (char)(65 + i); //Lettera maiuscola corrispondente al termine
                        sb.Append("~" + c.ToString()); //Negazione del termine
                    }
                    if (termini_essenziali[temp.IndexOf(s)][i] == '1') //Se uno viene lasciato così com'è
                    {
                        c = (char)(65 + i);
                        sb.Append(c.ToString());
                    }
                }
                sb.Append(" + ");
            }
            sb.Remove(sb.Length - 3, 3); //Rimuove l'ultimo " + " che rimane

            if (sb.ToString() == "f = ") //Se la funzione booleana è sempre vera viene restituita la stringa "f = 1"
            {
                sb.Append("1");
            }

            stato_finale.Add(sb.ToString());
            return stato_finale;
        }

        //Function: Stampa_Stato_Finale_Petrick
        //Funzione che attua il metodo di Petrick e restituisce la forma semplificata ottenuta
        public List<string> Stampa_Stato_Finale_Petrick()
        {
            List<string> stato_finale_petrick = new List<string>();
            string singola_riga = "";
            Ultimo_Passaggio();
            foreach (int n in termini_minimi) //Ricava una prima forma POS sulla base della tabella finale ottenuta con l'algoritmo Quine Mc-Cluskey
            {
                singola_riga += "(";
                foreach (string s in temp)
                {
                    foreach (string a in s.Split('-'))
                    {
                        int numero;
                        if (Int32.TryParse(a, out numero))
                            if (numero == n)
                            {
                                singola_riga = singola_riga + "P" + temp.IndexOf(s) + "+";
                            }
                    }
                }
                singola_riga = singola_riga.Remove(singola_riga.Length - 1, 1) + ")";
            }

            //Forma del tipo (P0+P1)(P0+P2)(P1+P3)(P2+P4)(P3+P5)(P4+P5)

            List<string> pos = new List<string>();
            List<string> sop = new List<string>();

            //Vengono separati i moltiplicandi della forma POS
            foreach (string s in singola_riga.Split(new[] { '(', ')' }, StringSplitOptions.RemoveEmptyEntries))
            {
                pos.Add(s);
            }

            Rimuovi_Duplicati(pos);
            string moltiplicazione = "";

            for (int i = 0; i < pos.Count(); i++)
            {
                moltiplicazione = Moltiplicazione(moltiplicazione, pos[i]); //Effettua una moltiplicazione progressiva
                moltiplicazione = Idempotenza_AND(moltiplicazione); //Seguita da una semplificazione tramite regole di Idempotenza
                moltiplicazione = Idempotenza_OR(moltiplicazione);
                moltiplicazione = Assorbimento(moltiplicazione); //E regola di assorbimento
            }

            pos.Clear(); //Viene liberato lo spazio in memoria allocato precedentemente per la forma POS
            //Vengono attuati gli ultimi passaggi di semplificazione sulla forma SOP finale ottenuta
            moltiplicazione = Idempotenza_AND(moltiplicazione);
            moltiplicazione = Idempotenza_OR(moltiplicazione);
            moltiplicazione = Assorbimento(moltiplicazione);
            moltiplicazione = moltiplicazione.Replace("(", "").Replace(")", "");

            foreach (string s in moltiplicazione.Split('+'))
            {
                sop.Add(s); //Separa gli addendi della forma SOP
            }

            //Viene calcolato il minor numero di termini possibile nella forma SOP
            int minimo = sop[0].Split(new[] { 'P' }, StringSplitOptions.RemoveEmptyEntries).Count();
            foreach (string s in sop)
            {
                string[] righe = s.Split(new[] { 'P' }, StringSplitOptions.RemoveEmptyEntries);
                if (righe.Count() <= minimo)
                {
                    minimo = righe.Count();
                }
            }

            List<string> sop_minime = new List<string>();
            foreach (string s in sop)
            {
                string[] sub = s.Split(new[] { 'P' }, StringSplitOptions.RemoveEmptyEntries);
                if (sub.Count() == minimo)
                {
                    sop_minime.Add(s);
                }
            }

            Valuta_Meno_Costosi(sop_minime); //Viene calcolato il costo letterale minimo assoluto

            char c;
            StringBuilder sb = new StringBuilder();

            //Alla fine vengono prese tutte le forme di costo letterale minimo equivalente e in modo simile alla procedura dell'algoritmo Quine McCluskey
            //Vengono riportate in forma SOP
            stato_finale_petrick.Add("Forme di costo letterale minimo equivalente ricavate tramite il metodo di Petrick:\n");

            foreach (string sop_minima in sop_minime)
            {
                sb.Clear();
                sb.Append("f = ");
                foreach (string s in sop_minima.Split(new[] { 'P' }, StringSplitOptions.RemoveEmptyEntries))
                {
                    for (int i = 0; i < termini_essenziali[Convert.ToInt32(s)].Length; i++)
                    {
                        if (termini_essenziali[Convert.ToInt32(s)][i] == '_')
                            sb.Append("");
                        if (termini_essenziali[Convert.ToInt32(s)][i] == '0') //Se zero viene negato
                        {
                            c = (char)(65 + i); //Lettera maiuscola corrispondente al termine
                            sb.Append("~" + c.ToString()); //Negazione del termine
                        }
                        if (termini_essenziali[Convert.ToInt32(s)][i] == '1') //Se uno viene lasciato così com'è
                        {
                            c = (char)(65 + i);
                            sb.Append(c.ToString());
                        }
                    }
                    sb.Append(" + ");
                }

                sb.Remove(sb.Length - 3, 3);
                if(sb.ToString() == "f = ")
                {
                    sb.Clear();
                    sb.Append("f = 1");
                }
                stato_finale_petrick.Add(sb.ToString());
                stato_finale_petrick.Add("\n\n");
            }
            return stato_finale_petrick;
        }
        //Function: Moltiplicazione
        //Procedura che restituisce il prodotto tra due moltiplicandi nella forma POS es. (P0+P1+P2)(P3+P4)
        private string Moltiplicazione(string a, string b) 
        {
            string risultato = "(";
            a = a.Replace("(", "");
            a = a.Replace(")", "");
            b = b.Replace("(", "");
            b = b.Replace(")", "");
            string[] moltiplicandi1 = a.Split('+');
            string[] moltiplicandi2 = b.Split('+');
            foreach (string s in moltiplicandi1)
            {
                foreach (string k in moltiplicandi2)
                {
                    risultato += (s + k);
                    risultato += "+";
                }
            }
            risultato = risultato.Remove(risultato.Length - 1, 1);
            risultato += ")";
            return risultato;
        }
        //Function: Idempotenza_AND
        //Procedura che effettua la semplificazione applicando la regola di idempotenza XX = X
        private string Idempotenza_AND(string mul)
        {
            string ris = "";
            List<string> termini1 = new List<string>();
            List<string> termini2 = new List<string>();
            List<string> termini3 = new List<string>();

            foreach (string s in mul.Split(new[] { '(', ')' }, StringSplitOptions.RemoveEmptyEntries))
            {
                termini1.Add(s);
            }
            Rimuovi_Duplicati(termini1);

            List<string> ordinamento = new List<string>();
            Ordinamento.Confronta_Termini cs = new Confronta_Termini();
            string espr = "";

            foreach (string s in termini1)
            {
                foreach (string k in s.Split(new[] { '+' }, StringSplitOptions.RemoveEmptyEntries))
                {
                    foreach(string j in k.Split(new[] { 'P' }, StringSplitOptions.RemoveEmptyEntries))
                    {
                        if (!ordinamento.Contains(j))
                        { ordinamento.Add(j); }
                    }

                    foreach (string j in ordinamento)
                    {
                        espr += "P" + j;
                    }

                    termini2.Add(espr);
                    espr = "";
                    ordinamento.Clear();
                }
                foreach(string k in termini2)
                {
                    espr += (k + "+");
                }
                termini3.Add(espr.Remove(espr.Length - 1, 1));
                espr = "";
                termini2.Clear();
            }


            foreach (string s in termini3)
            {
                ris = ris + "(" + s + ")";
            }
            return ris;
        }
        //Function: Idempotenza_OR
        //Procedura che effettua la semplificazione applicando la regola di idempotenza X + X = X
        private string Idempotenza_OR(string mul)
        {
            string ris = "";
            List<string> termini1 = new List<string>();
            List<string> termini2 = new List<string>();
            foreach (string s in mul.Split(new[] { '(', ')' }, StringSplitOptions.RemoveEmptyEntries))
            {
                termini1.Add(s);
            }
            Rimuovi_Duplicati(termini1); 
            List<string> sm = new List<string>();
            string espr = "";
            foreach (string s in termini1)
            {
                foreach (string k in s.Split(new[] { '+' }, StringSplitOptions.RemoveEmptyEntries))
                {
                    sm.Add(k);
                }
                Rimuovi_Duplicati(sm);

                foreach (string k in sm)
                {
                    espr += (k + "+");
                }

                espr = espr.Remove(espr.Length - 1, 1);
                termini2.Add(espr);

                espr = "";
                sm.Clear();
            }

            foreach (string s in termini2)
            {
                ris = ris + "(" + s + ")";
            }
            return ris;
        }
        //Function: Assorbimento
        //Procedura che effettua la semplificazione applicando la regola di assorbimento X + XY = X
        private string Assorbimento(string mul)
        {
            string ris = "";
            List<string> termini1 = new List<string>();
            foreach (string s in mul.Split(new[] { '+' }, StringSplitOptions.RemoveEmptyEntries))
            {
                termini1.Add(s);
            }
            Rimuovi_Duplicati(termini1);
            List<string> sm = new List<string>();
            foreach (string s in termini1)
            {
                foreach (string k in termini1)
                {
                    if (s.Split(new[] { 'P' }, StringSplitOptions.RemoveEmptyEntries).Count() == (k.Split(new[] { 'P' }, StringSplitOptions.RemoveEmptyEntries).Count() - 1))
                    {
                        List<string> sm1 = new List<string>();
                        List<string> sm2 = new List<string>();
                        int conto = 0;
                        foreach (string j in s.Split(new[] { 'P' }, StringSplitOptions.RemoveEmptyEntries))
                        {
                            sm1.Add(j);
                        }
                        foreach (string j in k.Split(new[] { 'P' }, StringSplitOptions.RemoveEmptyEntries))
                        {
                            sm2.Add(j);
                        }
                        foreach (string j in sm1)
                        {
                            foreach (string l in sm2)
                            {
                                if (j == l)
                                {
                                    conto++;
                                }
                            }
                        } 
                        if (conto == sm1.Count())
                        {
                            sm.Add(k);
                            conto = 0;
                        }
                        conto = 0;
                    }
                }
            }
            foreach (string s in sm)
            {
                termini1.Remove(s);
            }
            foreach (string s in termini1)
            {
                ris = ris + s + "+";
            }
            if(!String.IsNullOrEmpty(ris))
                return ris.Remove(ris.Length - 1, 1);
            return ris;
        }

        //Function: Valuta_Meno_Costosi
        // Valuta il costo letterale
        private void Valuta_Meno_Costosi(List<string> termini) 
        {
            StringBuilder sb = new StringBuilder();
            char c;
            int cl = (((numero_variabili + 1) * (int)Math.Pow(2, numero_variabili)) - (2 * numero_variabili)); //Il costo letterale massimo
            //che in qualunque caso è teoricamente possibile trovare
            int t_cl = 0;

            foreach(string s in termini)
            {
                foreach (string k in s.Split(new[] { 'P' }, StringSplitOptions.RemoveEmptyEntries))
                {
                    for (int i = 0; i < termini_essenziali[Convert.ToInt32(k)].Length; i++)
                    {
                        if (termini_essenziali[Convert.ToInt32(k)][i] == '_')
                            sb.Append("");
                        if (termini_essenziali[Convert.ToInt32(k)][i] == '0') //Se zero viene negato
                        {
                            c = (char)(65 + i); //Lettera maiuscola corrispondente al termine
                            sb.Append("~" + c.ToString()); //Negazione del termine
                        }
                        if (termini_essenziali[Convert.ToInt32(k)][i] == '1') //Se uno viene lasciato così com'è
                        {
                            c = (char)(65 + i);
                            sb.Append(c.ToString());
                        }
                    }
                    sb.Append(" + ");
                }

                sb.Remove(sb.Length - 3, 3);
                foreach (char car in sb.ToString())
                {
                    if (car >= 65 && car <= 90)
                        t_cl++;
                }

                if (t_cl <= cl)
                    cl = t_cl;

                t_cl = 0;
                sb.Clear();
            }

            for(int i = 0; i < termini.Count; i++)
            {
                foreach (string k in termini[i].Split(new[] { 'P' }, StringSplitOptions.RemoveEmptyEntries))
                {
                    for (int j = 0; j < termini_essenziali[Convert.ToInt32(k)].Length; j++)
                    {
                        if (termini_essenziali[Convert.ToInt32(k)][j] == '_')
                            sb.Append("");
                        if (termini_essenziali[Convert.ToInt32(k)][j] == '0') //Se zero viene negato
                        {
                            c = (char)(65 + j); //Lettera maiuscola corrispondente al termine
                            sb.Append("~" + c.ToString()); //Negazione del termine
                        }
                        if (termini_essenziali[Convert.ToInt32(k)][j] == '1') //Se uno viene lasciato così com'è
                        {
                            c = (char)(65 + j);
                            sb.Append(c.ToString());
                        }
                    }
                    sb.Append(" + ");
                }

                sb.Remove(sb.Length - 3, 3);
                foreach (char car in sb.ToString())
                {
                    if (car >= 65 && car <= 90)
                        t_cl++;
                }

                if(t_cl > cl)
                {
                    termini.Remove(termini[i]);
                }

                t_cl = 0;
                sb.Clear();
            }
        }
    }
}